@@ -4,14 +4,12 @@ use v5.26;
 
 package ChordPro;
 
-use App::Packager;
-
-use ChordPro::Version;
 use ChordPro::Utils;
 use ChordPro::Chords;
 use ChordPro::Output::Common;
 
-our $VERSION = $ChordPro::Version::VERSION;
+# Single line for stupid tools.
+use ChordPro::Version; our $VERSION = $ChordPro::Version::VERSION;
 
 =head1 NAME
 
@@ -75,6 +73,18 @@ our $config;
 
 package ChordPro;
 
+use ChordPro::Paths;
+use Encode qw(decode decode_utf8 encode_utf8);
+
+sub import {
+    # Add private library.
+    my $lib = CP->privlib;
+    for ( @INC ) {
+	return if $_ eq $lib;
+    }
+    unshift( @INC, $lib );
+}
+
 sub ::run {
     binmode(STDERR, ':utf8');
     binmode(STDOUT, ':utf8');
@@ -138,9 +148,7 @@ sub chordpro {
 	    # No default if more than one input document.
 	    die("Please use \"--output\" to specify the output file name\n");
 	}
-	my $f = $ARGV[0];
-	$f =~ s/\.\w+$/.pdf/;
-	$f .= ".pdf" if $f eq $ARGV[0];
+	my $f = CP->sibling( $ARGV[0], ext => ".pdf" );
 	$options->{output} = $f;
 	warn("Writing output to $f\n") if $options->{verbose};
     }
@@ -196,6 +204,9 @@ sub chordpro {
 		$opts->{meta}->{__config} = \@cfg;
 	    }
 	}
+	$opts->{generate} = $options->{generate};
+	# Wx runs on temp files, so pass real filename in.
+	$opts->{filesource} = $options->{filesource};
 	$s->parse_file( $file, $opts );
     }
 
@@ -274,19 +285,8 @@ sub chordpro {
 }
 
 sub ::dump {
-    use Data::Dumper qw();
-    local $Data::Dumper::Sortkeys  = 1;
-    local $Data::Dumper::Indent    = 1;
-    local $Data::Dumper::Quotekeys = 0;
-    local $Data::Dumper::Deparse   = 1;
-    local $Data::Dumper::Terse     = 1;
-    local $Data::Dumper::Trailingcomma = 1;
-    local $Data::Dumper::Useperl = 1;
-    local $Data::Dumper::Useqq     = 0; # I want unicode visible
-
-    my $s = Data::Dumper::Dumper @_;
-    defined wantarray or warn $s;
-    return $s;
+    use ChordPro::Dumper;
+    ddp(@_);
 }
 
 ################ Options and Configuration ################
@@ -704,37 +704,7 @@ sub app_setup {
     }
 
     # Config files.
-    if ( -d "/etc" ) {          # some *ux
-        $configs{sysconfig} =
-          File::Spec->catfile( "/", "etc", "$app_lc.json" );
-    }
-
-    if ( $ENV{XDG_CONFIG_HOME} && -d $ENV{XDG_CONFIG_HOME} ) {
-	$configs{userconfig} =
-	  File::Spec->catfile( $ENV{XDG_CONFIG_HOME}, $app_lc, "$app_lc.json" );
-	$ENV{CHORDPRO_LIB} ||= File::Spec->catfile( $ENV{XDG_CONFIG_HOME}, $app_lc);
-    }
-    elsif ( $ENV{HOME} && -d $ENV{HOME} ) {
-	my $dir = File::Spec->catfile( $ENV{HOME}, ".config" );
-        if ( -d $dir ) {
-            $configs{userconfig} =
-              File::Spec->catfile( $dir, $app_lc, "$app_lc.json" );
-	    $ENV{CHORDPRO_LIB} ||= File::Spec->catfile( $dir, $app_lc );
-        }
-        else {
-	    $dir = File::Spec->catfile( $ENV{HOME}, ".$app_lc" );
-            $configs{userconfig} =
-              File::Spec->catfile( $dir, "$app_lc.json" );
-	    $ENV{CHORDPRO_LIB} ||= $dir;
-        }
-    }
-
-    if ( -s ".$app_lc.json" ) {
-        $configs{config} = ".$app_lc.json";
-    }
-    else {
-        $configs{config} = "$app_lc.json";
-    }
+    %configs = %{ CP->configs };
 
     my $options =
       {
@@ -796,7 +766,7 @@ sub app_setup {
 
           ### Standard Chordii Options ###
 
-          "about|A" => \$about,         # About...
+          "about|A+" => \$about,        # About...
           "chord-font|C=s",             # Sets chord font
           "chord-grid-size|s=f",        # Sets chord diagram size [30]
           "chord-grids-sorted|S!",      # Prints chord diagrams ordered
@@ -900,8 +870,8 @@ sub app_setup {
         # Load Pod::Usage only if needed.
         require Pod::Usage;
         Pod::Usage->import;
-	my $f = $manual == 2 ? "pod/Config.pod" : "pod/ChordPro.pod";
-        unshift( @_, -input => getresource($f) );
+	my $f = $manual == 2 ? "Config.pod" : "ChordPro.pod";
+        unshift( @_, -input => CP->findres( $f, class => "pod" ) );
         &pod2usage;
     };
 
@@ -912,7 +882,7 @@ sub app_setup {
         $pod2usage->(VERBOSE => 2) if $manual;
     }
     app_ident(\*STDOUT, 0) if $version;
-    app_about(\*STDOUT, 0) if $about;
+    app_about(\*STDOUT, $about, 0) if $about;
 
     # If the user specified a config, it must exist.
     # Otherwise, set to a default.
@@ -933,11 +903,13 @@ sub app_setup {
         for ( $clo->{$config} ) {
             if ( defined($_) ) {
                 foreach my $c ( @$_ ) {
+		    my $try = $c;
 		    # Check for resource names.
-		    if ( $c !~ m;[/.]; ) {
-			$c = ::rsc_or_file( $c, "config" );
+		    if ( ! -r $try ) {
+			$try = CP->findcfg($c);
 		    }
-                    die("$c: $!\n") unless -r $c;
+                    die("$c: $!\n") unless $try && -r $try;
+		    $c = $try;
                 }
                 next;
             }
@@ -1025,8 +997,8 @@ sub app_ident {
 }
 
 sub app_about {
-    my ($fh, $exit) = @_;
-    print ${fh} <<EndOfAbout;
+    my ($fh, $level, $exit) = @_;
+    print ${fh} <<EndOfAbout,
 
 ChordPro: A lyrics and chords formatting program.
 
@@ -1041,91 +1013,130 @@ To learn more about ChordPro, look for the man page or do
 For more information, see https://www.chordpro.org .
 
 Run-time information:
-@{[::runtimeinfo()]}
 EndOfAbout
+      ::runtimeinfo( $level > 1 ? "extensive" : "normal" );
+
     exit $exit if defined $exit;
 }
 
-use Cwd qw(realpath);
+use List::Util qw(uniq);
 
 sub ::runtimeinfo {
-    my $short = shift;
-
-    my $fmt   = "  %-22.22s %s\n";
-    my $fmtv  = defined($Wx::VERSION) ? "  %s version %s\n" : $fmt;
-    my $fmtvv = defined($Wx::VERSION) ? "  %s %s\n" : $fmt;
+    my $level = shift // "normal";
+    my $cp = ChordPro::Paths->get;
+    my $fmt   = "  %-22.22s %-10s\n";
 
     # Sometimes version numbers are localized...
     my $dd = sub { my $v = $_[0]; $v =~ s/,/./g; $v };
 
-    my $msg = sprintf( $fmtv, "ChordPro core", $dd->($VERSION) );
+    my $msg = sprintf( $fmt, "ChordPro core", $dd->($VERSION) );
     $msg =~ s/core/reference/ if $::options->{reference};
     if ( $VERSION =~ /_/ ) {
 	$msg =~ s/\n$/ (Unsupported development snapshot)\n/;
     }
 
-    if ( $short ) {
+    if ( $level eq "short" ) {
 	$msg =~ s/^\s+//;
 	$msg =~ s/\s+/ /g;
 	$msg =~ s/\s*\n//;
 	return $msg;
     }
 
-    $msg .= sprintf( $fmtv, "Perl", $^V );
-    $msg =~ s/\n$/ ($^X)\n/;
-    if ( $App::Packager::PACKAGED ) {
-	my $p = App::Packager::Packager();
-	$p .= " Packager" unless $p =~ /packager/i;
-	$msg .= sprintf( $fmtv, $p, $dd->(App::Packager::Version()) );
-    }
+    $msg .= sprintf( $fmt, "Perl", $^V );
+    $msg =~ s/\n$/sprintf(" (%s)\n", $cp->display($^X))/e;
 
+    if ( CP->packager ) {
+	$msg .= sprintf( $fmt, CP->packager." Packager", CP->packager_version );
+    }
+ 
     # Determine resource path.
     my @p;
-    if ( $ENV{CHORDPRO_LIB} ) {
-	$msg .= sprintf( $fmtv, "CHORDPRO_LIB", $ENV{CHORDPRO_LIB} );
-	@p = splitpath($ENV{CHORDPRO_LIB});
+    my $tag = "CHORDPRO_LIB";
+    if ( defined $ENV{CHORDPRO_LIB} ) {
+	for ( $cp->path($ENV{CHORDPRO_LIB}) ) {
+	    $msg .= sprintf( $fmt, $tag, $cp->display($_) );
+	    $tag = "";
+	    push( @p, $cp->display($_) );
+	}
     }
-    push( @p, realpath( App::Packager::GetResourcePath() ) );
-    my $tag = "Resource path";
-    for ( @p ) {
-	$msg .= sprintf( $fmtvv, $tag, $_ );
+    push( @p,  map { $cp->display($_) } @{ $cp->resdirs } );
+    $tag = "Resource path";
+    for ( uniq(@p) ) {
+	$msg .= sprintf( $fmt, $tag, $_ );
 	$tag = "";
     }
+    eval { require ChordPro::Delegate::ABC;
+	   my $x;
+	   if ( ChordPro::Delegate::ABC::have_xs() ) {
+	       $x = ChordPro::Delegate::ABC::packaged_qjs();
+	       $msg .= sprintf( $fmt, "ABC support",
+				$x->[0] . " (" . $x->[1] . ")" );
+	   }
+	   elsif ( $x = findexe( "abc2svg", "silent" )
+		        || findexe( "abcnode", "silent" ) ) {
+	       $msg .= sprintf( $fmt, "ABC support", $cp->display($x) );
+	   }
+	   elsif ( $x = ChordPro::Delegate::ABC::packaged_qjs() ) {
+	       $msg .= sprintf( $fmt, "ABC support",
+				$cp->display($x->[0]) . " (" . $cp->display($x->[-1]) . ")" );
+	   }
+    };
 
+    my $vv = sub {
+	my ( $mod ) = @_;
+	no strict 'refs';
+	$msg .= sprintf( $fmt, $mod, $dd->(${${"${mod}::"}{VERSION}}) );
+	return unless $level eq "extensive";
+	chomp($msg);
+	my $pm = $mod =~ s;::;/;gr . ".pm";
+	my $loc = $INC{$pm};
+	if ( 0 ) {
+	    $msg .= "(site)" if $loc =~ /site_perl/;
+	    $msg .= "(vendor)" if $loc =~ /vendor_perl/;
+	    $msg .= "(ChordPro)" if $loc =~ /ChordPro\/lib/;
+	    $msg .= "(private)" if $loc =~ /$ENV{HOME}\/lib\/perl5/;
+	}
+	else {
+	    $msg .= "($1)" if $loc =~ /^(.*)\/\Q$pm\E$/;
+	}
+	$msg .= "\n";
+    };
     $msg .= "\nModules and libraries:\n";
     if ( defined $Wx::VERSION ) {
 	no strict 'subs';
-	$msg .= sprintf( $fmtv, "wxPerl", $dd->($Wx::VERSION) );
-	$msg .= sprintf( $fmtv, "wxWidgets", $dd->(Wx::wxVERSION) );
+	$msg .= sprintf( $fmt, "wxPerl", $dd->($Wx::VERSION) );
+	$msg .= sprintf( $fmt, "wxWidgets", $dd->(Wx::wxVERSION) );
     }
 
     local $SIG{__WARN__} = sub {};
     local $SIG{__DIE__} = sub {};
-    $msg .= sprintf( $fmtv, "Storable", $dd->($Storable::VERSION) );
-    $msg .= sprintf( $fmtv, "Object::Pad", $dd->($Object::Pad::VERSION) );
+    $vv->("Storable");
+    $vv->("Object::Pad");
     eval { require Text::Layout;
-	$msg .= sprintf( $fmtv, "Text::Layout", $dd->($Text::Layout::VERSION) );
+	$vv->("Text::Layout");
     };
     eval { require HarfBuzz::Shaper;
-	$msg .= sprintf( $fmtv, "HarfBuzz::Shaper", $dd->($HarfBuzz::Shaper::VERSION) );
-	$msg .= sprintf( $fmtv, "HarfBuzz library", $dd->(HarfBuzz::Shaper::hb_version_string()) );
+	$vv->("HarfBuzz::Shaper");
+	$msg .= sprintf( $fmt, "HarfBuzz library", $dd->(HarfBuzz::Shaper::hb_version_string()) );
     };
-    $msg .= sprintf( $fmtv, "File::LoadLines", $dd->($File::LoadLines::VERSION) );
+    $vv->("File::LoadLines");
     eval { require PDF::Builder;
-	$msg .= sprintf( $fmtv, "PDF::Builder", $dd->($PDF::Builder::VERSION) );
-    }
-    or
+	$vv->("PDF::Builder");
+    };
     eval { require PDF::API2;
-	$msg .= sprintf( $fmtv, "PDF::API2", $dd->($PDF::API2::VERSION) );
+	$vv->("PDF::API2");
+	eval { require PDF::API2::XS;
+	       $vv->("PDF::API2::XS");
+	};
+    };
+    eval { require SVGPDF;
+	$vv->("SVGPDF");
     };
     eval { require Font::TTF;
-	$msg .= sprintf( $fmtv, "Font::TTF", $dd->($Font::TTF::VERSION) );
+	$vv->("Font::TTF");
     };
-    eval { require Image::Magick;
-	$msg .= sprintf( $fmtv, "Image::Magick",
-			 $dd->( $Image::Magick::VERSION ||
-				$Image::Magick::Q16::VERSION ||
-				$Image::Magick::Q8::VERSION || "6.x?" ) );
+    eval { require JavaScript::QuickJS;
+	$vv->("JavaScript::QuickJS");
     };
     return $msg;
 }
@@ -1222,40 +1233,6 @@ EndOfUsage
     exit $exit if defined $exit;
 }
 
-################ Resources ################
-
-use Encode qw(decode decode_utf8 encode_utf8);
-
-sub ::rsc_or_file {
-    my ( $c, $cfg ) = @_;
-    my $f = $c;
-    $cfg .= "/" if $cfg;
-
-    # Check for resource names.
-    if ( $f !~ m;[/.]; ) {
-	if ( $c =~ /^(.+):(.*)/ ) {
-	    $f = $cfg . lc($1) . "/" . lc($2) . ".json";
-	}
-	else {
-	    $f = $cfg . lc($c) . ".json";
-	}
-    }
-    if ( $ENV{CHORDPRO_LIB} ) {
-	my @libs = splitpath($ENV{CHORDPRO_LIB});
-	foreach my $lib ( @libs ) {
-	    $lib = expand_tilde($lib);
-	    warn("RSC1: $lib/$f\n") if $options->{debug};
-	    return $lib . "/" . $f if -r $lib . "/" . $f;
-	}
-    }
-
-    warn("RSC3: $f\n") if $options->{debug};
-    my $t = getresource($f);
-    return defined($t) ? $t : $c;
-}
-
-use lib ( grep { defined } getresource("CPAN") );
-
 =head1 FONTS
 
 There are two ways to specify fonts: with a font filename, and a
